"""
.. _tut-acc-standard-weighted-average-methods-pipeline-sim:

##################################################################
ACC Standard analysis with different average methods (Simulated)
##################################################################

In this example we simulate an ACC response and average it using standard average as well as two different weighted
average method.
The first method, estimates the weights by tracking the same time points across epochs. This means that a number of
epochs needs to be obtained to estimate the weights of those epochs.
The second method estimates the weights within each epoch. This means that each epoch has its own weight.
For cortical responses, the latter can be more beneficial as artefacts within an epoch will be outweighed for that epoch
only.

"""
# Enable below for interactive backend
# import matplotlib
# if 'Qt5Agg' in matplotlib.rcsetup.all_backends:
#     matplotlib.use('Qt5Agg')
from peegy.processing.pipe.pipeline import PipePool
from peegy.processing.pipe.simulate import GenerateInputData
from peegy.processing.tools.detection.definitions import TimePeakWindow, PeakToPeakMeasure, TimeROI
from peegy.processing.tools.template_generator.auditory_waveforms import aep
from peegy.processing.pipe.general import FilterData, ReSampling, AutoRemoveBadChannels, ReferenceData, \
    BaselineCorrection
from peegy.processing.pipe.epochs import EpochData, AverageEpochs
from peegy.processing.pipe.statistics import HotellingT2Test
from peegy.processing.pipe.detection import PeakDetectionTimeDomain
from peegy.processing.pipe.plot import PlotWaveforms, PlotTopographicMap
from peegy.processing.pipe.attach import AppendGFPChannel
from peegy.processing.pipe.storage import MeasurementInformation, SubjectInformation, SaveToDatabase
import os
import numpy as np
import astropy.units as u

# %%
# Time windows
# ============================
# Time windows are used to define where to look for different peaks


tw = np.array([TimePeakWindow(ini_time=50 * u.ms, end_ref='N1', label='P1', positive_peak=True,
                              exclude_channels=['GFP']),
               TimePeakWindow(ini_time=80 * u.ms, end_time=200 * u.ms, label='N1', positive_peak=False,
                              exclude_channels=['GFP']),
               TimePeakWindow(ini_ref='N1', end_time=300 * u.ms, label='P2', positive_peak=True,
                              exclude_channels=['GFP']),
               TimePeakWindow(ini_time=50 * u.ms, end_time=150 * u.ms, label='gfp1', positive_peak=True,
                              target_channels=['GFP']),
               TimePeakWindow(ini_time=150 * u.ms, end_time=500 * u.ms, label='gfp2', positive_peak=True,
                              target_channels=['GFP'])])

# %%
# Peak-to-peak Measures
# ============================
# Peak-to-peak measures are defined based on the labels of the TimePeaks defined above.


pm = np.array([PeakToPeakMeasure(ini_peak='N1', end_peak='P2')])

# %%
# Time regions of interest
# ============================
# TimeROI are defined as time regions where different measures will be performed, e.g. SNR measure or statistical
# measures


roi_windows = np.array(
    [TimeROI(ini_time=100.0 * u.ms, end_time=250.0 * u.ms, window_label="acc_snr"),
     TimeROI(ini_time=500.0 * u.ms, end_time=600.0 * u.ms, window_label="control")])

# %%
# Generate some data
# ===================
# First we generate some ACC data

n_channels=32
fs = 512.0 * u.Hz
template_waveform, _ = aep(fs=fs)
event_times = np.arange(0, 100.0, 1.0) * u.s
reader = GenerateInputData(template_waveform=template_waveform,
                           fs=fs,
                           n_channels=n_channels,
                           mixing_matrix=np.diag(np.arange(n_channels))/n_channels,
                           layout_file_name='biosemi32.lay',
                           snr=0.1,
                           event_times=event_times,
                           event_code=1.0,
                           figures_subset_folder='acc_test_weighted_average_types',
                           include_eog_events=True
                           )
reader.run()

# %%
# Start the pipeline
# ============================
# Now we proceed with our basic processing pipeline


pipeline = PipePool()
pipeline['referenced'] = ReferenceData(reader,
                                       reference_channels=['Cz'],
                                       invert_polarity=True)
pipeline['channel_cleaned'] = AutoRemoveBadChannels(pipeline.referenced)
pipeline['down_sampled'] = ReSampling(pipeline.channel_cleaned,
                                      new_sampling_rate=256. * u.Hz)
# %%
# Continue with the pipeline
# ------------------------------------

pipeline['time_filtered_data'] = FilterData(pipeline['down_sampled'],
                                            high_pass=2.0 * u.Hz,
                                            low_pass=30.0 * u.Hz)

pipeline.run()

# %%
# Get Epochs
# ------------------------------------
# We partition the data into epochs or trials based on the event code used.


pipeline['time_epochs'] = EpochData(pipeline.time_filtered_data,
                                    event_code=1.0)
pipeline.run()

# %%
# Compute global field power (GFP)
# ------------------------------------
# We compute the GFP across channels and epochs. A new channel with the GFP is appended to the data

pipeline['time_epochs_with_gfp'] = AppendGFPChannel(pipeline['time_epochs'])

# %%
# Base line correction
# ------------------------------------
# We correct the initial segment by removing the mean in the first 20 ms
pipeline['time_epochs_with_baseline_corrected'] = BaselineCorrection(
    pipeline['time_epochs_with_gfp'],
    ini_time=0 * u.ms,
    end_time=20 * u.ms)

pipeline.run()

# %%
# Compute HT2 statistics
# ------------------------------------
# Using the raw epochs we estimate Hotelling-T2 statistics

pipeline['ht2_standard'] = HotellingT2Test(pipeline.time_epochs_with_baseline_corrected,
                                           roi_windows=roi_windows,
                                           weight_data=False)

pipeline['ht2_across'] = HotellingT2Test(pipeline.time_epochs_with_baseline_corrected,
                                         roi_windows=roi_windows,
                                         weight_data=True,
                                         weight_across_epochs=True)
pipeline['ht2_within'] = HotellingT2Test(pipeline.time_epochs_with_baseline_corrected,
                                         roi_windows=roi_windows,
                                         weight_data=True,
                                         weight_across_epochs=False)

pipeline.run()

# %%
# Compute average responses
# ------------------------------------
# We compute weighted average on epochs with and without spatial filtering (DSS)

pipeline['time_average_standard'] = AverageEpochs(pipeline['time_epochs_with_baseline_corrected'],
                                                  roi_windows=roi_windows,
                                                  weighted_average=False)
pipeline['time_average_standard_gfp'] = AppendGFPChannel(pipeline['time_average_standard'],
                                                         channel_label='GFP2')


pipeline['time_average_across'] = AverageEpochs(pipeline['time_epochs_with_baseline_corrected'],
                                                roi_windows=roi_windows,
                                                weighted_average=True,
                                                weight_across_epochs=True)
pipeline['time_average_across_gfp'] = AppendGFPChannel(pipeline['time_average_across'],
                                                       channel_label='GFP2')

pipeline['time_average_within'] = AverageEpochs(pipeline['time_epochs_with_baseline_corrected'],
                                                roi_windows=roi_windows,
                                                weighted_average=True,
                                                weight_across_epochs=False)
pipeline['time_average_within_gfp'] = AppendGFPChannel(pipeline['time_average_within'],
                                                       channel_label='GFP2')

# %%
# Detect peaks
# ------------------------------------
# Using the average data, we proceed detecting peaks

pipeline['data_with_peaks_standard'] = PeakDetectionTimeDomain(
    pipeline['time_average_standard_gfp'],
    time_peak_windows=tw,
    peak_to_peak_measures=pm)


pipeline['data_with_peaks_across'] = PeakDetectionTimeDomain(
    pipeline['time_average_across_gfp'],
    time_peak_windows=tw,
    peak_to_peak_measures=pm)

pipeline['data_with_peaks_within'] = PeakDetectionTimeDomain(
    pipeline['time_average_within_gfp'],
    time_peak_windows=tw,
    peak_to_peak_measures=pm)

# %%
# Show some waveforms
# ------------------------------------


pipeline['plotter'] = PlotWaveforms(pipeline['data_with_peaks_standard'],
                                    ch_to_plot=np.array(['T7', 'C4', 'CP2', 'GFP']),
                                    overlay=[pipeline['data_with_peaks_across'], pipeline['data_with_peaks_within']],
                                    statistical_test='f_test_time',
                                    show_following_stats=['f', 'rn'],
                                    return_figures=True,
                                    )
pipeline.run()

# %%
# Show some topographic maps
# ------------------------------------

pipeline['topographic_map_1'] = PlotTopographicMap(pipeline['data_with_peaks_standard'],
                                                   topographic_channels=np.array(['T7', 'C4', 'CP2', 'GFP']),
                                                   plot_x_lim=[0, 0.8],
                                                   plot_y_lim=[-3, 3],
                                                   return_figures=True)


pipeline['topographic_map_1'] = PlotTopographicMap(pipeline['data_with_peaks_across'],
                                                   topographic_channels=np.array(['T7', 'C4', 'CP2', 'GFP']),
                                                   plot_x_lim=[0, 0.8],
                                                   plot_y_lim=[-3, 3],
                                                   return_figures=True)
pipeline['topographic_map_2'] = PlotTopographicMap(pipeline['data_with_peaks_within'],
                                                   topographic_channels=np.array(['T7',     'C4', 'CP2', 'GFP']),
                                                   plot_x_lim=[0, 0.8],
                                                   plot_y_lim=[-3, 3],
                                                   return_figures=True)
pipeline.run()

# now we save our data to a database
subject_info = SubjectInformation(subject_id='Test_Subject')
measurement_info = MeasurementInformation(
    date='Today',
    experiment='sim')

_parameters = {'Type': 'ACC'}
database_path = reader.input_node.paths.test_path.joinpath('acc_test_data_weighted_average_types.sqlite')
pipeline['database'] = SaveToDatabase(database_path=database_path,
                                      measurement_information=measurement_info,
                                      subject_information=subject_info,
                                      recording_information={'recording_device': 'dummy_device'},
                                      stimuli_information=_parameters,
                                      processes_list=[pipeline['data_with_peaks_standard'],
                                                      pipeline['data_with_peaks_across'],
                                                      pipeline['data_with_peaks_within'],
                                                      pipeline['ht2_standard'],
                                                      pipeline['ht2_across'],
                                                      pipeline['ht2_within']]
                                      )
pipeline.run()

# %%
# Generate pipeline diagram
# ------------------------------------
pipeline.diagram(file_name=reader.output_node.paths.figures_current_dir + 'pipeline.png',
                 return_figure=True,
                 dpi=600)



from peegy.processing.pipe.definitions import InputOutputProcess
from peegy.processing.tools.epochs_processing_tools import et_mean
import multiprocessing
from peegy.plot import eeg_ave_epochs_plot_tools as eegpt
from peegy.definitions.channel_definitions import Domain
import pyfftw
import numpy as np
import os
import astropy.units as u
from PyQt5.QtCore import QLibraryInfo
from ssqueezepy import Wavelet, cwt, stft
from ssqueezepy.experimental import scale_to_freq
os.environ["QT_QPA_PLATFORM_PLUGIN_PATH"] = QLibraryInfo.location(QLibraryInfo.PluginsPath)


class HilbertEnvelope(InputOutputProcess):
    def __init__(self, input_process=InputOutputProcess, high_pass=None, low_pass=None, **kwargs):
        """
        This process computes the  Hilbert Envelope of EEG data
        :param input_process: InputOutputProcess Class
        :param kwargs: extra parameters to be passed to the superclass
        """
        super(HilbertEnvelope, self).__init__(input_process=input_process, **kwargs)

    def transform_data(self):
        data = self.input_node.data.copy()
        _fft = pyfftw.builders.fft(data, overwrite_input=False, planner_effort='FFTW_ESTIMATE', axis=0,
                                   threads=multiprocessing.cpu_count())
        fx = _fft()
        n = fx.shape[0]
        h = np.zeros(n)
        if n % 2 == 0:
            h[0] = h[n // 2] = 1
            h[1:n // 2] = 2
        else:
            h[0] = 1
            h[1:(n + 1) // 2] = 2
        _ifft = pyfftw.builders.ifft(fx * h.reshape(-1, 1), overwrite_input=False, planner_effort='FFTW_ESTIMATE',
                                     axis=0,
                                     threads=multiprocessing.cpu_count())
        hilbert_data = _ifft()
        self.output_node.data = np.abs(hilbert_data)


class InducedResponse(InputOutputProcess):
    def __init__(self, input_process=InputOutputProcess,
                 weighted_average=True,
                 n_tracked_points=256,
                 block_size=5,
                 roi_windows=None, **kwargs):
        super(InducedResponse, self).__init__(input_process=input_process, **kwargs)
        self.weighted_average = weighted_average
        self.n_tracked_points = n_tracked_points
        self.block_size = block_size
        self.roi_windows = roi_windows

    def transform_data(self):
        trials_abs_w_ave, w, rn, cum_rn, w_fft, *_ = \
            et_mean(epochs=np.abs(self.input_node.data),
                    block_size=max(self.block_size, 5),
                    samples_distance=int(max(self.input_node.data.shape[0] // self.n_tracked_points, 10)),
                    roi_windows=self.roi_windows,
                    weighted=self.weighted_average
                    )
        w_ave, w, rn, cum_rn, w_fft, *_ = \
            et_mean(epochs=self.input_node.data,
                    block_size=max(self.block_size, 5),
                    samples_distance=int(max(self.input_node.data.shape[0] // self.n_tracked_points, 10)),
                    roi_windows=self.roi_windows,
                    weighted=self.weighted_average
                    )
        self.output_node.data = trials_abs_w_ave - np.abs(w_ave)
        self.output_node.rn = rn
        self.output_node.cum_rn = cum_rn
        self.output_node.snr = None
        self.output_node.cum_snr = None
        self.output_node.s_var = None


class TimeFrequencyResponse(InputOutputProcess):
    def __init__(self, input_process=InputOutputProcess,
                 time_window: u.Quantity = 0.3 * u.s,
                 sample_interval: u.Quantity = 0.004 * u.s,
                 topographic_channels=np.array([]),
                 title: str = '',
                 plot_x_lim=None,
                 plot_y_lim=None,
                 times=np.array([]),
                 fig_format='.png',
                 fontsize=12,
                 user_naming_rule: str = '',
                 spec_thresh=6,
                 average_mode='magnitude',
                 method: str = 'wavelet',
                 wavelet_type: str = 'morlet',
                 **kwargs):
        super(TimeFrequencyResponse, self).__init__(input_process=input_process, **kwargs)
        self.time_window = time_window
        self.sample_interval = sample_interval
        self.frequency = None
        self.time = None
        self.topographic_channels = topographic_channels
        self.title = title
        self.plot_x_lim = plot_x_lim
        self.plot_y_lim = plot_y_lim
        self.fig_format = fig_format
        self.fontsize = fontsize
        self.user_naming_rule = user_naming_rule
        self.times = times
        self.time_window = time_window
        self.sample_interval = sample_interval
        self.spec_thresh = spec_thresh
        self.average_mode = average_mode
        self.method = method
        self.wavelet_type = wavelet_type

    def transform_data(self):
        magnitude, freqs = self.time_frequency_transformation(epochs=self.input_node.data,
                                                              fs=self.input_node.fs,
                                                              time_window=self.time_window,
                                                              sample_interval=self.sample_interval,
                                                              average_mode=self.average_mode,
                                                              method=self.method,
                                                              wavelet_type=self.wavelet_type
                                                              )
        self.output_node.data = magnitude
        self.output_node.y = freqs
        self.output_node.x = self.input_node.x
        self.output_node.domain = Domain.time_frequency

    @staticmethod
    def time_frequency_transformation(epochs: np.array = np.array([]),
                                      method='spectrogram',
                                      wavelet_type='morlet',
                                      time_window=2.0,
                                      sample_interval=2.0,
                                      fs=0.0,
                                      average_mode='magnitude'):

        _waves, freqs, ave = None, None, None
        data = epochs
        if data.ndim == 2:
            data = data[:, :, None]
        if average_mode == 'complex':
            data = np.mean(epochs, 2, keepdims=True)
        for _i in range(data.shape[2]):
            if method == 'wavelet':
                wavelet = Wavelet(wavelet_type)
                w, scales = cwt(data[:, :, _i].T, wavelet)
                freqs = scale_to_freq(scales, wavelet, w.shape[2], fs=fs.to(u.Hz).value)
                _waves = w
            if method == 'stft':
                stft_data = stft(data[:, :, _i].T)[:, ::-1, :]
                freqs = np.linspace(1, 0, stft_data.shape[1]) * fs.to(u.Hz).value / 2
                _waves = stft_data
            if average_mode == 'magnitude':
                _waves = np.abs(_waves)
            if _i == 0:
                ave = _waves
            else:
                ave += _waves
            print('Computing spectrogram {:} out of {:}'.format(_i, data.shape[2]))
        ave /= data.shape[2]
        ave = np.transpose(ave, [2, 0, 1])
        freqs = freqs * u.Hz
        ave = ave * data.unit
        return ave, freqs


class PlotTimeFrequencyData(InputOutputProcess):
    def __init__(self, input_process=InputOutputProcess,
                 topographic_channels=np.array([]),
                 title: str = '',
                 plot_x_lim: [float, float] = None,
                 plot_y_lim: [float, float] = None,
                 times: np.array = np.array([]),
                 fig_format: str = '.png',
                 fontsize: float = 12,
                 user_naming_rule: str = '',
                 spec_thresh: float = -np.inf,
                 return_figures: bool = True,
                 save_figures: bool = False,
                 normalize: bool = True,
                 db_scale: bool = True,
                 **kwargs):
        super(PlotTimeFrequencyData, self).__init__(input_process=input_process, **kwargs)
        self.topographic_channels = topographic_channels
        self.title = title
        self.plot_x_lim = plot_x_lim
        self.plot_y_lim = plot_y_lim
        self.fig_format = fig_format
        self.fontsize = fontsize
        self.user_naming_rule = user_naming_rule
        self.times = times
        self.spec_thresh = spec_thresh
        self.return_figures = return_figures
        self.save_figures = save_figures
        self.normalize = normalize
        self.db_scale = db_scale

    def transform_data(self):
        assert self.input_node.domain == Domain.time_frequency, 'input should be a time-frequency transformed data'

        figures = eegpt.plot_eeg_time_frequency_power(
            ave_data=self.input_node,
            eeg_topographic_map_channels=self.topographic_channels,
            figure_dir_path=self.input_node.paths.figures_current_dir,
            figure_basename='{:}{:}'.format(
                self.input_process.name + '_',
                self.user_naming_rule),
            title=self.title,
            x_lim=self.plot_x_lim,
            y_lim=self.plot_y_lim,
            fig_format=self.fig_format,
            fontsize=self.fontsize,
            spec_thresh=self.spec_thresh,
            return_figures=self.return_figures,
            save_figures=self.save_figures,
            normalize=self.normalize,
            db_scale=self.db_scale
        )
        self.figures = figures



import peegy.io.readers.edf_bdf_reader as br
import matplotlib.pyplot as plt
import numpy as np
# import bids.layout
import os
__author__ = 'jundurraga'

test_frequencies = np.array([6.8, 20.4, 40.9])
_path = os.path.abspath(os.path.dirname(__file__))
folder_name = os.path.join(_path, "../test_data/set_1/")
# layout = bids.layout.BIDSLayout(folder_name)
# f_name = layout.get(type='eeg', extensions='bdf')[0].filename
f_name = ''
header = br.read_edf_bdf_header(file_name=f_name)
ch = []
[ch.append(header['channels'][x]) for x in [0]]
data, *_ = br.get_data(header=header, channels=ch, ini_time=0, end_time=None)
plt.plot(data)
plt.show()



from peegy.processing.events.event_tools import events_to_samples_array
from peegy.definitions.events import Events
from peegy.processing.tools.filters.resampling import eeg_resampling
import numpy as np
import logging
import astropy.units as u
import soundfile as sf
log = logging.getLogger()


def data_to_wav(data: np.array(u.Quantity) = None,
                events: Events | None = None,
                output_file_name: str | None = None,
                fs: type(u.Quantity) | None = None,
                fs_wav: type(u.Quantity) | None = None,
                normalize: bool = False,
                gain: float = 0.99):

    data_resampled, _factor = eeg_resampling(x=data,
                                             new_fs=fs_wav,
                                             fs=fs)

    events_samples = events_to_samples_array(events=events, fs=fs_wav, n_samples=data_resampled.shape[0])
    if normalize:
        _data = data_resampled / np.max(np.abs(data_resampled), axis=0)
    else:
        _data = data_resampled
    _data = gain * _data.value
    if np.max(np.abs(_data)) >= 1:
        print('your file is clipping')
    events_samples = events_samples / np.max(np.abs(events_samples))
    _data = np.hstack((_data, events_samples))
    sf.write(file=output_file_name,
             data=_data,
             samplerate=int(fs_wav.to('Hz').value))
    return _data, fs_wav



"""
.. _tut-assr-using-own-data-trials-test-sim:

#################################################################################
ASSR F-Test Example using 'external' data (time x channels x trials) (Simulated)
#################################################################################

This example illustrates how you can use any generic data (e.g. matlab data, or text file data) in the shape of time x
channels x trials.
In this example we simulate an ASSR, and we investigate the significance of the response by means of an F-test that
compares the variance of the frequency of interest with the variance of neighbouring frequency bins

"""
# Enable below for interactive backend
# import matplotlib
# if 'qt5agg' in matplotlib.backends.backend_registry.list_builtin():
#     matplotlib.use('qt5agg')
from peegy.processing.pipe.pipeline import PipePool
from peegy.processing.pipe.definitions import Domain
from peegy.processing.pipe.io import GenericInputData
from peegy.processing.pipe.attach import AppendGFPChannel
from peegy.processing.pipe.statistics import FTest
from peegy.processing.pipe.general import FilterData
from peegy.processing.pipe.epochs import AverageEpochs
from peegy.processing.pipe.plot import PlotWaveforms, PlotTopographicMap
from peegy.processing.pipe.spatial_filtering import CreateAndApplySpatialFilter
from peegy.processing.pipe.storage import MeasurementInformation, SubjectInformation, SaveToDatabase
from peegy.processing.pipe.attach import CombineChannels
import peegy.definitions.default_paths as default_paths
import os
import astropy.units as u
import numpy as np

# %%
# Generate some data (time x channels x trials)
# =============================================
# We generate some data in the shape of time x channels x trials and put it into a GenericInputData class

fs = 256.0 * u.Hz
epoch_length = 4.0 * u.s
epoch_length = np.ceil(epoch_length * fs) / fs  # fit to fs rate
assr_frequency = np.array([41.0]) * u.Hz
assr_frequency = np.ceil(epoch_length * assr_frequency) / epoch_length  # fit to epoch length
# here we generate and pass our own data
n_epochs = 20
time = np.arange(0, epoch_length.to(u.s).value, 1 / fs.to(u.Hz).value) * u.s
my_sin = np.sin(2 * np.pi * u.rad * assr_frequency * time)
template_waveform = np.array([my_sin, 0.5 * my_sin, 0.25 * my_sin]).T
template_waveform = np.repeat(template_waveform[:, :, None], n_epochs, axis=2)
# add some noise
np.random.seed(1)
template_waveform = template_waveform + np.random.randn(*template_waveform.shape)

reader = GenericInputData(data=template_waveform,
                          fs=fs,
                          figures_folder=default_paths.test_figures_path,
                          figures_subset_folder='my_own_trials_assr_f_test')
reader.run()

# %%
# Start the pipeline
# ============================
pipeline = PipePool()
pipeline['time_filtered_data'] = FilterData(reader,
                                            high_pass=2.0 * u.Hz,
                                            low_pass=100.0 * u.Hz)
pipeline.run()

# %%
# Get DSS components for cleaned data
# ------------------------------------
# Compute spatial filter based on artefact free epochs

pipeline['dss_time_epochs'] = CreateAndApplySpatialFilter(pipeline['time_filtered_data'],
                                                          test_frequencies=np.concatenate((assr_frequency,
                                                                                           2 * assr_frequency)),
                                                          components_to_plot=np.arange(0, 5),
                                                          projection_domain=Domain.frequency)
pipeline.run()


pipeline['combined'] = CombineChannels(pipeline['time_filtered_data'],
                                       channel_labels=['CH_0', 'CH_1'],
                                       function=lambda x: np.mean(x, axis=1, keepdims=True)
                                       )

# %%
# Compute global field power (GFP)
# ------------------------------------
# We compute the GFP across channels and epochs. A new channel with the GFP is appended to the data

pipeline['dss_time_epochs_with_gfp'] = AppendGFPChannel(pipeline['dss_time_epochs'])
pipeline['time_epochs_with_gfp'] = AppendGFPChannel(pipeline['combined'])
pipeline.run()

# %%
# Compute average responses
# ------------------------------------
# Apply averages to data

pipeline['time_ave_dss'] = AverageEpochs(pipeline['dss_time_epochs_with_gfp'])
pipeline['time_ave'] = AverageEpochs(pipeline['time_epochs_with_gfp'])
pipeline.run()

# %%
# Run F-Test
# ------------------------------------
# Now we run an F-test to determine if frequency component is significant

pipeline['fft_ave_f_tests_dss'] = FTest(pipeline['time_ave_dss'],
                                        n_fft=int(epoch_length*fs*1),
                                        test_frequencies=assr_frequency,
                                        delta_frequency=9.0 * u.Hz)
pipeline['fft_ave_f_tests'] = FTest(pipeline['time_ave'],
                                    n_fft=int(epoch_length * fs * 1),
                                    test_frequencies=assr_frequency,
                                    delta_frequency=9.0 * u.Hz)
pipeline.run()

# %%
# Generate figures
# ------------------------------------
# Now plot the results


pipeline['topographic_map_1'] = PlotTopographicMap(pipeline['fft_ave_f_tests_dss'],
                                                   topographic_channels=np.array(['CH_0', 'CH_1', 'CH_2']),
                                                   plot_x_lim=[0, 90],
                                                   plot_y_lim=[0, 4],
                                                   return_figures=True,
                                                   user_naming_rule='dss')
pipeline['topographic_map_2'] = PlotTopographicMap(pipeline['fft_ave_f_tests'],
                                                   topographic_channels=np.array(['CH_0', 'CH_1', 'CH_2']),
                                                   plot_x_lim=[0, 90],
                                                   plot_y_lim=[0, 4],
                                                   return_figures=True,
                                                   user_naming_rule='standard')
pipeline['plotter'] = PlotWaveforms(pipeline['fft_ave_f_tests'],
                                    ch_to_plot=np.array(['CH_0', 'CH_1', 'CH_2', 'CH_0_CH_1']),
                                    overlay=[pipeline['fft_ave_f_tests_dss']],
                                    plot_x_lim=[0, 90],
                                    statistical_test='f_test_freq',
                                    show_following_stats=['frequency_tested', 'f'],
                                    user_naming_rule='waveforms',
                                    return_figures=True,
                                    fig_format='.png')

pipeline.run()

# %%
# Get generated data and save to database
# ------------------------------------------
# We get the measurements we are interested in and save them into a database

subject_info = SubjectInformation(subject_id='Test_Subject')
measurement_info = MeasurementInformation(
    date='Today',
    experiment='sim')
_parameters = {'Type': 'ASSR'}
database_path = reader.input_node.paths.test_path.joinpath('assr_own_trials_data_f_test_data.sqlite')
pipeline['database'] = SaveToDatabase(database_path=database_path,
                                      measurement_information=measurement_info,
                                      subject_information=subject_info,
                                      recording_information={'recording_device': 'dummy_device'},
                                      stimuli_information=_parameters,
                                      processes_list=[pipeline['fft_ave_f_tests'],
                                                      pipeline['fft_ave_f_tests_dss']],
                                      include_waveforms=True
                                      )
pipeline.run()

# %%
# Generate pipeline diagram
# ------------------------------------
pipeline.diagram(file_name=reader.output_node.paths.figures_current_dir + 'pipeline.png',
                 return_figure=True,
                 dpi=600)



from peegy.processing.pipe.definitions import InputOutputProcess
from peegy.processing.tools.epochs_processing_tools import et_subtract_oeg_template, et_subtract_correlated_ref
from peegy.processing.tools.template_generator.auditory_waveforms import fade_in_out_window
from peegy.processing.tools.multiprocessing.multiprocessesing_filter import filt_data
from peegy.processing.pipe.epochs import EpochData, AverageEpochs
import matplotlib.pyplot as plt
import peegy.processing.tools.filters.eegFiltering as eegf
from peegy.processing.tools.filters.resampling import eeg_resampling
from peegy.processing.tools.eeg_epoch_operators import et_unfold
import numpy as np
from sklearn.decomposition import FastICA
from sklearn.cross_decomposition import CCA
from sklearn.neighbors import KernelDensity
from peegy.processing.tools.math_tools import get_local_minima
import itertools
import os
import astropy.units as u
from peegy.tools.units.unit_tools import set_default_unit
from PyQt5.QtCore import QLibraryInfo
from tqdm import tqdm
import copy
from typing import List
os.environ["QT_QPA_PLATFORM_PLUGIN_PATH"] = QLibraryInfo.location(QLibraryInfo.PluginsPath)


class ReSampling(InputOutputProcess):
    def __init__(self, input_process=InputOutputProcess,
                 new_sampling_rate: u.quantity.Quantity | None = None,
                 **kwargs):
        """
        This InputOutputProcess class will resample the data to the new_sampling_rate
        :param input_process: InputOutputProcess Class
        :param new_sampling_rate: float indicating the new sampling rate
        :param kwargs: extra parameters to be passed to the superclass
        """
        super(ReSampling, self).__init__(input_process=input_process, **kwargs)
        self.new_sampling_rate = set_default_unit(new_sampling_rate, u.Hz)

    def transform_data(self):
        _data = self.input_node.data.copy()
        data, _factor = eeg_resampling(x=_data,
                                       new_fs=self.new_sampling_rate,
                                       fs=self.input_node.fs)
        # data = resample(self.input_node.data.copy(), data.shape[0], t=None, axis=0, window=None, domain='time')
        # self.output_node.data = data * _data.unit
        self.output_node.data = data
        self.output_node.fs = self.input_node.fs * _factor
        self.output_node.process_history.append(self.process_parameters)
        self.output_node.statistical_tests = self.input_process.output_node.statistical_tests


class ReferenceData(InputOutputProcess):
    def __init__(self, input_process=InputOutputProcess,
                 reference_channels: List[str] | None = None,
                 remove_reference=True,
                 invert_polarity=False,
                 **kwargs):
        """
        This class will reference the data by subtracting the reference_channels (mean) from each individual channel
        :param input_process: InputOutputProcess Class
        :param reference_channels: list of string with the reference channels labels, if empty across channel average
        will be used
        :param remove_reference: boolean indicating whether to keep or remove the reference channel from the data
        :param invert_polarity: boolean indicating if channels polarity will be polarity inverted (data * -1)
        :param kwargs: extra parameters to be passed to the superclass
        """
        super(ReferenceData, self).__init__(input_process=input_process, **kwargs)
        if reference_channels is None:
            reference_channels = ['']
        self.reference_channels = reference_channels
        self.remove_reference = remove_reference
        self.invert_polarity = invert_polarity

    def transform_data(self):
        _ch_idx = self.input_node.get_channel_idx_by_label(self.reference_channels)
        if not _ch_idx.size:
            self.remove_reference = False
            _ch_idx = np.arange(self.input_node.layout.size)
        print('Referencing data to: ' + ''.join(['{:s} '.format(_ch.label) for _ch in self.input_node.layout[_ch_idx]]))
        reference = np.mean(self.input_node.data[:, _ch_idx], axis=1, keepdims=True)
        self.output_node.data = (self.input_node.data - reference) * (-1.0) ** self.invert_polarity

        if self.remove_reference:
            self.output_node.delete_channel_by_idx(_ch_idx)


class FilterData(InputOutputProcess):
    def __init__(self, input_process=InputOutputProcess,
                 high_pass: u.quantity.Quantity | None = None,
                 low_pass: u.quantity.Quantity | None = None,
                 zero_latency: bool = True,
                 ripple_db: u.Quantity = 60 * u.dB,
                 transition_width: u.Quantity = 1 * u.Hz,
                 **kwargs):
        """
        Filter EEG data using a zero group-delay technique
        :param input_process: InputOutputProcess Class
        :param high_pass: Frequency (in Hz) of high-pass filter
        :param low_pass: Frequency (in Hz) of low-pass filter
        :param zero_latency: If true, the latency (group delay) of the filter will be adjusted by (N + 1) / 2 which is
        the group delay of a linear phase FIR filter
        :param transition_width: width of filter transition band (in Hz)
        :param ripple_db: amount of ripple, in dB, in the pass-band and rejection regions. The magnitude variation will
        be below -ripple dB within the bandpass region, whilst the attenuation in the rejection band will be at least
        + ripple dB
        :param kwargs: extra parameters to be passed to the superclass
        """
        super(FilterData, self).__init__(input_process=input_process, **kwargs)
        self.low_pass = set_default_unit(low_pass, u.Hz)
        self.high_pass = set_default_unit(high_pass, u.Hz)
        self.zero_latency = zero_latency
        self.ripple_db = ripple_db
        self.transition_width = transition_width

    def transform_data(self):
        if self.low_pass is None and self.high_pass is None:
            self.output_node.data = self.input_node.data
            return
        filtered_signal = self.input_node.data.copy()
        _b = eegf.bandpass_fir_win(high_pass=self.high_pass,
                                   low_pass=self.low_pass,
                                   fs=self.input_node.fs,
                                   ripple_db=self.ripple_db.value,
                                   transition_width=self.transition_width,
                                   filt_filt_cutoff=False)
        _b = _b * u.dimensionless_unscaled
        if self.zero_latency:
            mode = 'original'
        else:
            mode = 'full'
        filtered_signal = filt_data(data=filtered_signal, b=_b, mode=mode)
        self.output_node.data = filtered_signal


class RegressOutEOG(InputOutputProcess):
    def __init__(self, input_process=InputOutputProcess,
                 ref_channel_labels: List[str] | None = None,
                 method: str = 'template',
                 high_pass: u.quantity.Quantity = 0.1 * u.Hz,
                 low_pass: u.quantity.Quantity = 20.0 * u.Hz,
                 peak_width: u.quantity.Quantity = 0.1 * u.s,
                 template_width: u.quantity.Quantity = 1.4 * u.s,
                 remove_eog_channels: bool = True,
                 save_figure: bool = True,  # save figure
                 fig_format: str = '.png',
                 user_naming_rule: str = '',
                 return_figures: bool = False,
                 n_iterations: int = 10,
                 kernel_bandwidth: float = 4 * 0.15,
                 use_initial_template: bool = True,
                 **kwargs):
        """
        This class removes EOG artifacts using a template technique. Blinking artifacts are detected and averaged to
        generate a template. This template is scaled for each channel in order to maximize correlation between each
        individual blink and individual events on each channel
        The resulting template is removed from the data, reducing blinks artifacts.
        :param input_process: InputOutputProcess Class
        :param ref_channel_labels: a list with the channel labels that contain the EOG
        :param high_pass: Frequency (in Hz) of high-pass filter all data. This is necessary to generate the template
        :param low_pass: Frequency (in Hz) of low-pass filter only applied to EOG channels
        :param peak_width: default minimum width (in seconds) to detect peaks
        :param template_width: the duration (in seconds) of the time window to average and generate a template
        :param remove_eog_channels: if true EOG channels will be removed once data has been cleaned
        :param save_figure: whether to save or not figures showing the detection and removal of blinks
        :param fig_format: format of output figure
        :param n_iterations: number of iterations to improve the template estimation
        :param kernel_bandwidth: factor use to control the with of the Gaussian kernel in threshold detection
        matching.
        :param use_initial_template: if true, a template eye blinking will be used as starting point to find events
        :param kwargs: extra parameters to be passed to the superclass
        """
        super(RegressOutEOG, self).__init__(input_process=input_process, **kwargs)
        self.ref_channel_labels = ref_channel_labels
        self.low_pass = low_pass
        self.high_pass = high_pass
        self.peak_width = peak_width
        self.template_width = template_width
        self.remove_eog_channels = remove_eog_channels
        self.save_figure = save_figure
        self.fig_format = fig_format
        self.method = method
        self.user_naming_rule = user_naming_rule
        self.return_figures = return_figures
        self.n_iterations = n_iterations
        self.kernel_bandwidth = kernel_bandwidth
        self.use_initial_template = use_initial_template

    def transform_data(self):
        data = self.input_node.data.copy()
        _ref_idx = self.input_node.get_channel_idx_by_label(labels=self.ref_channel_labels)
        figure_dir_path = self.input_node.paths.figures_current_dir
        _sep = '_' if self.user_naming_rule is not None else ''
        _naming_rule = '' if self.user_naming_rule is None else self.user_naming_rule
        figure_basename = self.name + _sep + _naming_rule
        artefact_method = self.method
        if artefact_method is None:
            if data.ndim == 3:
                artefact_method = 'correlation'
            else:
                artefact_method = 'template'

        if _ref_idx.size:
            print('Using {:} to remove eog artifacts'.format(artefact_method))
            if artefact_method == 'template':
                data, figures, *_ = et_subtract_oeg_template(data=data,
                                                             idx_ref=np.array(_ref_idx),
                                                             high_pass=self.high_pass,
                                                             low_pass=self.low_pass,
                                                             fs=self.input_node.fs,
                                                             template_width=self.template_width,
                                                             plot_results=self.save_figure,
                                                             figure_path=figure_dir_path,
                                                             figure_basename=figure_basename,
                                                             n_iterations=self.n_iterations,
                                                             kernel_bandwidth=self.kernel_bandwidth,
                                                             use_initial_template=self.use_initial_template
                                                             )
            if artefact_method == 'correlation':
                data, figures = et_subtract_correlated_ref(data=data,
                                                           idx_ref=np.array(_ref_idx),
                                                           high_pass=self.high_pass,
                                                           low_pass=self.low_pass,
                                                           fs=self.input_node.fs,
                                                           plot_results=self.save_figure,
                                                           figure_path=figure_dir_path,
                                                           figure_basename=figure_basename)

            if self.return_figures:
                self.figures = figures
            else:
                [plt.close(fig) for fig in figures]
        else:
            print('Reference channels for eye artefact reduction are not valid! Will return data unaltered.')

        self.output_node.data = data
        if self.remove_eog_channels and _ref_idx.size:
            self.output_node.delete_channel_by_idx(_ref_idx)


class RegressOutICA(InputOutputProcess):
    def __init__(self, input_process=InputOutputProcess,
                 ref_channel_labels: List[str] | None = None,
                 **kwargs):
        """
        This class uses independent component analysis to remove EOG activity. The removal is based on the correlation
        between the ICA and the reference channels (containing large EOG activity)
        :param input_process: InputOutputProcess Class
        :param ref_channel_labels: a list with the channel labels that contain the EOG
        :param kwargs:
        """
        super(RegressOutICA, self).__init__(input_process=input_process, **kwargs)
        self.ref_channel_labels = ref_channel_labels

    def transform_data(self):
        data = self.input_node.data.copy()
        _ref_idx = self.input_node.get_channel_idx_by_label(labels=self.ref_channel_labels)
        _cov = data[:, _ref_idx]
        _offset = int(data.shape[0] * 0.10)

        # Compute ICA
        print('Performing ICA analysis')
        ica = FastICA(n_components=data.shape[1], max_iter=1000)
        ica.fit(data)
        # decompose signal into components
        components = ica.fit_transform(data)
        corr_coefs = np.empty((components.shape[1], _cov.shape[1]))
        for _i_com, _i_cov in itertools.product(range(components.shape[1]), range(_cov.shape[1])):
            corr_coefs[_i_com, _i_cov] = \
                np.corrcoef(_cov[_offset:-_offset, _i_cov], components[_offset:-_offset, _i_com])[0, 1]
        _idx_to_remove = np.argmax(np.abs(corr_coefs), axis=0)
        print('Maximum correlations: {:}'.format(corr_coefs[np.argmax(np.abs(corr_coefs), axis=0), :]))
        print('Removing components: {:}'.format(_idx_to_remove))
        components[:, _idx_to_remove] = 0
        clean_data = ica.inverse_transform(components)
        self.output_node.data = clean_data


class RegressOutCCA(InputOutputProcess):
    def __init__(self, input_process=InputOutputProcess, ref_channel_labels: List[str] | None = None, **kwargs):
        super(RegressOutCCA, self).__init__(input_process=input_process, **kwargs)
        self.channel_labels = ref_channel_labels

    def transform_data(self):
        data = self.input_node.data.copy()
        _ref_idx = self.input_node.get_channel_idx_by_label(labels=self.ref_channel_labels)
        _cov = data[:, _ref_idx]
        # Compute ICA
        print('Performing ICA analysis')

        cca = CCA(n_components=data.shape[1])
        cca.fit(data, _cov)
        # decompose signal into components
        X_c, Y_c = cca.transform(data, _cov)
        # corr_coefs = np.empty((components.shape[1], _cov.shape[1]))
        # for _i_com, _i_cov in itertools.product(range(components.shape[1]), range(_cov.shape[1])):
        #     corr_coefs[_i_com, _i_cov] = \
        #     np.corrcoef(_cov[_offset:-_offset, _i_cov], components[_offset:-_offset, _i_com])[0, 1]
        # _idx_to_remove = np.argmax(np.abs(corr_coefs), axis=0)
        # print('Maximum correlations: {:}'.format(corr_coefs[np.argmax(np.abs(corr_coefs), axis=0), :]))
        # print('Removing components: {:}'.format(_idx_to_remove))
        # components[:, _idx_to_remove] = 0
        # clean_data = ica.inverse_transform(components)

        # self.output_node.data = clean_data


class AutoRemoveBadChannels(InputOutputProcess):
    def __init__(self, input_process=InputOutputProcess,
                 thr_sd=5.0 * u.dimensionless_unscaled,
                 amp_thr=50000.0 * u.uV,
                 interval=0.001 * u.s,
                 **kwargs):
        """
        This function will try to detect bad channels by looking at the standard deviation across channels.
        It will remove any channels with and std larger than thr_sd, which is computed across all channels.
        It also removes any channel whose amplitude exceeds amp_thr.
        :param input_process: InputOutputProcess Class
        :param thr_sd: threshold standard deviation to remove channels. Channels with larger std will be removed
        :param amp_thr: threshold amplitude. Channel exceeding this will be removed
        :param interval: time interval to subsample the data before estimating std (this is used to speed up the process
        in very long data files
        :param kwargs: extra parameters to be passed to the superclass
        """
        super(AutoRemoveBadChannels, self).__init__(input_process=input_process, **kwargs)
        self.thr_sd = thr_sd
        self.amp_thr = set_default_unit(amp_thr, u.uV)
        self.interval = set_default_unit(interval, u.s)

    def transform_data(self):
        if self.input_node.data.ndim == 3:
            data = et_unfold(self.input_node.data)
        else:
            data = self.input_node.data
        step_size = np.maximum(int(self.interval * self.input_node.fs), 1)
        _samples = np.arange(0, self.input_node.data.shape[0], step_size).astype(int)
        sub_data = data[_samples, :]
        # compute dc component
        _dc_component = np.mean(np.abs(sub_data), axis=0)
        bad_channels = np.where(_dc_component > self.amp_thr)[0]
        _others_idx = np.array([idx for idx in np.arange(sub_data.shape[1]) if idx not in bad_channels], dtype=int)
        a_std = np.std(sub_data[:, _others_idx], axis=0)
        thr_ci = self.thr_sd * np.std(a_std) + np.mean(a_std)
        n_ch_idx = np.where(a_std > thr_ci)[0]
        bad_idx = _others_idx[n_ch_idx] if n_ch_idx.size else np.array([], dtype=int)
        bad_channels = np.concatenate((bad_channels, bad_idx))
        _bad_channels_index = np.unique(bad_channels)
        self.output_node.data = self.input_node.data.copy()
        self.output_node.delete_channel_by_idx(_bad_channels_index)


class RemoveBadChannels(InputOutputProcess):
    def __init__(self, input_process=InputOutputProcess, bad_channels: List[str] | None = None, **kwargs):
        """
        This class will remove any channel passed in bad_channel
        :param input_process: InputOutputProcess Class
        :param bad_channels: list of strings with the label of the channels to be removed
        :param kwargs: extra parameters to be passed to the superclass
        """
        super(RemoveBadChannels, self).__init__(input_process=input_process, **kwargs)
        if bad_channels is None:
            bad_channels = ['']
        self.bad_channels = bad_channels

    def transform_data(self):
        idx_bad_channels = [_i for _i, _item in enumerate(self.input_node.layout) if _item.label in self.bad_channels]
        _bad_channels_index = np.unique(idx_bad_channels)
        self.output_node.data = self.input_node.data.copy()
        if _bad_channels_index.size:
            self.output_node.delete_channel_by_idx(_bad_channels_index)


class BaselineCorrection(InputOutputProcess):
    def __init__(self, input_process=InputOutputProcess,
                 ini_time: u.quantity.Quantity | None = None,
                 end_time: u.quantity.Quantity | None = None,
                 **kwargs):
        """
        This class will remove the mean of the data on each row between the specified times (ini_time and end_time)
        :param input_process: InputOutputProcess Class
        :param ini_time: initial time (in sec) from which mean will be calculated
        :param end_time: end time (in sec) for which the mean will be calculated
        :param kwargs: extra parameters to be passed to the superclass
        """
        super(BaselineCorrection, self).__init__(input_process=input_process, **kwargs)
        self.ini_time = set_default_unit(ini_time, u.s)
        self.end_time = set_default_unit(end_time, u.s)

    def transform_data(self):
        data = self.input_node.data
        ini_time = 0.0 * u.s if self.ini_time is None else self.ini_time
        end_time = np.inf * u.s if self.end_time is None else self.end_time
        _samples = self.input_node.x_range_to_samples(ini_time=ini_time, end_time=end_time)
        data = data - np.mean(data[_samples], axis=0)
        self.output_node.data = data


class SetBadEvents(InputOutputProcess):
    def __init__(self, input_process=InputOutputProcess,
                 reject_dc: bool = True,
                 rejection_percentage: float | None = None,
                 rejection_threshold: u.quantity.Quantity | None = None,
                 window_duration: u.quantity.Quantity | None = None,
                 max_percentage_epochs_above_threshold: float = 1.0,
                 event_code: float | None = None,
                 events_mask: int | None = None,
                 kernel_bandwidth: float = 0.15,
                 **kwargs):
        """
        This class will epochs where a given threshold has been exceeded. If any channel exceeds the threshold at a
        given trial, that particular epoch will be removed from all channels. This is done to keep the data in a single
        matrix.
        :param input_process: InputOutputProcess Class
        :param reject_dc: bool indicating if detected DC should be found and marked as bad events
        :param max_percentage_epochs_above_threshold: if a single channel has more than max_epochs_above_threshold
        (percentage) epochs, the channel will be removed.
        :param rejection_percentage indicates the percentage of epochs to remove.
        above the threshold, the channel will be removed.
        :param rejection_threshold: indicates level above which epochs will be rejected
        :param kwargs: extra parameters to be passed to the superclass
        """
        super(SetBadEvents, self).__init__(input_process=input_process, **kwargs)
        self.reject_dc = reject_dc
        self.rejection_threshold = set_default_unit(rejection_threshold, u.uV)
        self.event_code = event_code
        self.events_mask = events_mask
        self.kernel_bandwidth = kernel_bandwidth
        self.window_duration = window_duration

    def transform_data(self):
        if not self.input_node.events.events().size:
            print('no events in data')
            return
        data = self.input_node.data
        self.output_node.events = copy.deepcopy(self.input_node.events)
        self.output_node.data = data
        self.output_node.events.mask = self.events_mask

        _time = self.window_duration
        if self.window_duration is None:
            _time = np.mean(np.diff(self.output_node.events.get_events_time(code=self.event_code)))
        _window_samples = int(_time * self.input_node.fs)
        events_index = self.output_node.events.get_events_index(code=self.event_code,
                                                                fs=self.input_node.fs)
        _derivarive = np.diff(data, axis=0)
        _idx_bad_events_dc = np.array([], dtype=int)
        if self.reject_dc:
            std = np.array([]) * data.unit
            for i, _event_ini in enumerate(tqdm(events_index, desc='Finding DC intervals')):
                # ensure that blocks match buffer size
                _event_ini = np.minimum(np.maximum(0, _event_ini), _derivarive.shape[0])
                _event_end = np.maximum(0,
                                        np.minimum(_event_ini + _window_samples, _derivarive.shape[0]))
                _current_buffer = _derivarive[_event_ini:_event_end, :]
                if _current_buffer.size == 0:
                    continue
                _std = np.std(_current_buffer, axis=0, keepdims=True)
                if i == 0:
                    std = _std
                else:
                    std = np.vstack((std, _std))
            if std.size > 0:
                q3, q1 = np.percentile(std, [95, 5], axis=0)
                iqr = q3 - q1

                bw = iqr * self.kernel_bandwidth
                if np.all(bw.value != 0):
                    for idx_ch in range(data.shape[1]):
                        kde = KernelDensity(bandwidth=bw[idx_ch].value, kernel='gaussian')
                        kde.fit(std[:, idx_ch][:, None].value)
                        x_d = np.linspace(std[:, idx_ch].min() - 5 * bw[idx_ch],
                                          std[:, idx_ch].max() + 5 * bw[idx_ch], 1000).reshape(-1, 1)
                        log_dens = kde.score_samples(x_d.value)
                        _pdf = np.exp(log_dens)
                        _minima, _minima_idx = get_local_minima(_pdf.squeeze())
                        _max_idx = _pdf.argmax()
                        threshold = None
                        if _minima_idx[_minima_idx < _max_idx].size:
                            threshold_idx = _minima_idx[_minima_idx < _max_idx][-1]
                            threshold = x_d[threshold_idx]
                            _idx_bad_events_dc = np.append(_idx_bad_events_dc, np.argwhere(std[:, idx_ch] < threshold))
                        # if show_histogram:
                        #     fig = plt.figure()
                        #     ax = fig.add_subplot(111)
                        #     ax.fill(x_d, _pdf, fc="#AAAAFF")
                        #     ax.plot(x_d[threshold_idx], _pdf[threshold_idx], 'o')
                        #     ax.plot(x_d[_max_idx], _pdf[_max_idx], 'o')
                        #     fig.show()
        _idx_bad_events_thr = np.array([], dtype=int)
        if self.rejection_threshold is not None:
            for i, _event_ini in enumerate(tqdm(events_index,
                                                desc='Finding data intervals above {:}'.format(
                                                    self.rejection_threshold))):
                # ensure that blocks match buffer size
                _event_ini = np.minimum(np.maximum(0, _event_ini), data.shape[0])
                _event_end = np.maximum(0,
                                        np.minimum(_event_ini + _window_samples, data.shape[0]))
                _current_buffer = data[_event_ini:_event_end, :]
                if _current_buffer.size == 0:
                    continue
                if np.any(np.abs(_current_buffer) > self.rejection_threshold):
                    _idx_bad_events_thr = np.append(_idx_bad_events_thr, i)

        _idx_bad_events = np.unique(np.concatenate((_idx_bad_events_dc, _idx_bad_events_thr)))
        if _idx_bad_events.size:
            self.output_node.events.set_events_status(event_idx=_idx_bad_events,
                                                      bad_event=True)
            print('{:} out of {:} events set as bad based on DC precense'.format(
                _idx_bad_events_dc.size,
                events_index.size
            ))
            print('{:} out of {:} events set as bad based on rejection threshold criteria of {:}'.format(
                _idx_bad_events_thr.size,
                events_index.size,
                self.rejection_threshold
            ))
            print('Setting {:} out of {:} events as bad (based on threshold and DC)'.format(
                _idx_bad_events.size,
                events_index.size
            ))
            print(self.output_node.events.summary())


class RemoveEvokedResponse(InputOutputProcess):
    def __init__(self, input_process=InputOutputProcess,
                 pre_stimulus_interval: u.quantity.Quantity | None = None,
                 post_stimulus_interval: u.quantity.Quantity | None = None,
                 baseline_correction_ini_time: u.quantity.Quantity | None = None,
                 baseline_correction_end_time: u.quantity.Quantity | None = None,
                 event_code: float | None = None,
                 base_line_correction: bool = False,
                 demean: bool = True,
                 events_mask: int | None = None,
                 weighted_average: bool = True,
                 weight_across_epochs: bool = False,
                 n_tracked_points: int | None = None,
                 block_size: int = 1,
                 fade_in_out: bool = False,
                 fade_in_out_percentage: float = 0.05,
                 **kwargs):
        """
        This class will remove the evoked response from the continuous data and return the continuous data.
        :param input_process: InputOutputProcess Class
        :param pre_stimulus_interval: the length (in sec) of the data to be read before the trigger
        :param post_stimulus_interval: the length (in sec) of the data to be read after the trigger
        :param event_code: integer indicating the event code to be used to epoch the data
        :param base_line_correction: whether to perform baseline correction or not. The initial and final time points
        used to compute the baseline correction are given by baseline_correction_ini_time and
        baseline_correction_end_time
        :param baseline_correction_ini_time: initial time (relative to trigger) to compute the mean for baseline
        correction. For example, if trigger is at zero, and you want to compute the baseline correction from -200 ms you
        can set baseline_correction_ini_time = -200 * u.ms
        :param baseline_correction_end_time: end time (relative to trigger) to compute the mean for baseline
        correction. For example, if trigger is at zero, and you want to compute the baseline correction
        between -200 * ms and -50 ms, you can set baseline_correction_ini_time = -200 * u.ms and
        baseline_correction_end_time = -50 * u.ms
        :param demean: whether to remove the mean from each epoch
        :param events_mask: integer value used to masker triggers codes. This is useful to ignore triggers inputs above
        a particular value. For example, if only the first 8 trigger inputs were used (max decimal value is 255), in a
        system with 16 trigger inputs, then the masker could be set to 255 to ignore any trigger from trigger inputs 9
        to 16
        :param weighted_average: if True, weighted average will be used
        :param weight_across_epochs: if True, weights are computed across epochs (as in Elbeling 1984) otherwise weights
        are computed within epoch (1 / variance across time)
        :param n_tracked_points: number of equally spaced points over time used to estimate residual noise and weights
        :param block_size: number of trials that will be stacked together to estimate the residual noise
        :param fade_in_out: It True, a faded in and out window will be applied to evoked response to prevent edge
        artefacts
        :param fade_in_out_percentage: defents the percentage of the signal in which fade-in and fade-out will take
        place. E.g., if epoch is 1.0s, then the fade-in and fade-out will last 50 ms if fade_in_out_percentage is 0.05
        :param kwargs: extra parameters to be passed to the superclass
        """
        super(RemoveEvokedResponse, self).__init__(input_process=input_process, **kwargs)
        self.pre_stimulus_interval = set_default_unit(pre_stimulus_interval, u.s)
        self.post_stimulus_interval = set_default_unit(post_stimulus_interval, u.s)
        self.event_code = event_code
        self.base_line_correction = base_line_correction
        self.demean = demean
        self.events_mask = events_mask
        self.baseline_correction_ini_time = baseline_correction_ini_time
        self.baseline_correction_end_time = baseline_correction_end_time
        self.weighted_average = weighted_average
        self.n_tracked_points = n_tracked_points
        self.block_size = block_size
        self.weight_across_epochs = weight_across_epochs
        self.fade_in_out = fade_in_out
        self.fade_in_out_percentage = fade_in_out_percentage

    def transform_data(self):
        data = self.input_node.data.copy()
        epochs = EpochData(input_process=self.input_process,
                           pre_stimulus_interval=self.pre_stimulus_interval,
                           post_stimulus_interval=self.post_stimulus_interval,
                           event_code=self.event_code,
                           events_mask=self.events_mask,
                           base_line_correction=self.base_line_correction,
                           baseline_correction_ini_time=self.baseline_correction_ini_time,
                           baseline_correction_end_time=self.baseline_correction_end_time,
                           keep_input_node=True
                           )
        epochs.run()
        evoked_response = AverageEpochs(input_process=epochs,
                                        keep_input_node=False,
                                        weighted_average=self.weighted_average,
                                        weight_across_epochs=self.weight_across_epochs,
                                        block_size=self.block_size,
                                        n_tracked_points=self.n_tracked_points)
        evoked_response.run()
        _epoch_samples = evoked_response.output_node.data.shape[0]
        evoked_data = evoked_response.output_node.data
        if self.fade_in_out:
            fade_window = fade_in_out_window(n_samples=_epoch_samples,
                                             fraction=self.fade_in_out_percentage)
            evoked_data = evoked_data * fade_window

        for _event_idx in self.input_node.events.get_events_index(self.event_code, fs=self.input_node.fs):
            _ini = _event_idx
            _end = np.minimum(_event_idx + _epoch_samples, data.shape[0])
            _b_size = _end - _ini
            if _b_size > 0:
                _er = evoked_data[0: _end - _ini, :]
                data[_ini: _end, :] = data[_ini: _end, :] - _er
        self.output_node.data = data



import matplotlib.pyplot as plt
from peegy.io.eeg.reader import eeg_reader
from peegy.io.exporters.edf_bdf_writer import bdf_resampler
from peegy.io.external_tools.file_tools import get_files_and_meta_files
import numpy as np
from peegy.processing.events.event_tools import get_events
import os
from pathlib import Path
import matplotlib
import astropy.units as u
if 'Qt5Agg' in matplotlib.rcsetup.all_backends:
    matplotlib.use('Qt5Agg')


_path = Path(os.path.abspath(os.path.dirname(__file__)))
data_folder = _path.parent.parent.absolute().joinpath("test_data/")
eeg_data_directory = r'/home/jundurraga/Documents/Measurements/peegy_data/IPM-FR/'
data_files_meta = get_files_and_meta_files(eeg_data_directory, file_types=['bdf'])

for _, _file in data_files_meta.iterrows():
    output_file_path = _path.parent.parent.absolute().joinpath('split_test')
    _file_path = _file.data_links.data_file
    bdf_resampler(input_file_name=_file_path,
                  # output_path=output_file_path,
                  new_fs=1024 * u.Hz,
                  trash_original=True)

data_files_meta = get_files_and_meta_files(eeg_data_directory, file_types=['bdf'])
_first_file = data_files_meta.iloc[0].data_links.data_file
data4 = eeg_reader(file_name=_first_file)
data_ds, events_ds, _, _ = data4.get_data(channels_idx=np.array([0]))
et_ds = get_events(event_channel=events_ds, fs=data4.fs)
et_ds.summary()
fig1 = plt.figure(figsize=(7, 6), dpi=100)
t_ds = np.arange(data_ds.shape[0]) / data4.fs
ax = fig1.add_subplot(1, 1, 1)
ax.plot(t_ds, data_ds[:, 0])
ax.plot(t_ds, events_ds)
plt.show()



import numpy as np
import astropy.units as u


def set_default_unit(value: type(np.array) | None = None,
                     unit=u.Unit) -> u.quantity.Quantity:
    """
    This function will convert a numpy array into a Quantity using the default unit, only if the input does not have
    a unit
    :param value: array to set default units
    :param unit: the default unit to use
    :return: array with units
    """
    if value is None:
        return
    if isinstance(value, u.quantity.Quantity):
        out = value
    else:
        out = value * unit
        print('No metric unit set, assuming {:}'.format(unit))
    return out



import pandas as pd
from peegy.processing.pipe.definitions import InputOutputProcess
from peegy.io.storage.data_storage_tools import store_data, MeasurementInformation, SubjectInformation, PandasDataTable
from peegy.processing.statistics.definitions import TestType
import time
from typing import List


class SaveToDatabase(InputOutputProcess):
    def __init__(self,
                 database_path: str | None = None,
                 subject_information: SubjectInformation | None = None,
                 measurement_information: MeasurementInformation | None = None,
                 recording_information: dict | None = None,
                 stimuli_information: dict | None = None,
                 processes_list: List[InputOutputProcess] | None = None,
                 include_waveforms: bool = False,
                 names_prefixes: List[str] | None = None,
                 check_duplicates: bool = True):
        """
        This InputOutputProcess class saves the required data into a sqlite database that can be used to perform large
        analysis, for example, reading the database in R.
        :param database_path: full path and name where database will be created and saved (e.g. /home/user/data.sqlite)
        :param subject_information: class filled with the required subject information
        :param measurement_information: : class filled with the required measurement information
        :param recording_information: standard dictionary containing keys and values with useful information of
        recording device (e.g. {'sampling_rate': 16384, 'device': 'biosemi'})
        :param stimuli_information: standard dictionary containing keys and values with useful information of
        stimulus used for current data (e.g. {'amplitude': 0.5, 'stimulus': 'sinusoidal', 'carrier_frequency': 500})
        :param processes_list: a list of InputOutputProcess from which statistical tables, amplitudes and peaks will
        be extracted and saved.
        :param include_waveforms: boolean or list of boolean indicating if waveforms from each InputOutputProcess should
        be saved in the database. If a boolean variable is passed, that value will apply to all processes.
        If a list is passed, it must be the same size as process_list. Only the waveforms of the process_list item
        corresponding to the same index of include_waveform element == True will be saved.
        :param names_prefixes: either None or a list of strings of the same length as processes_list. When empty,
        tables created will have as prefix the name of the InputOutputProcess. If passes, tables will have the prefix
        passed in this list.
        :param check_duplicates: If true, entries from statistical tables contained by the process passed in the
        processes_list will be checked and replaced if a match exists.
        If false, repeated entries will be allowed. This may be useful to speedup saving data on a new database where
        no duplicate checks are necessary, but not recommendable when adding data into an existing database.

        """

        super(SaveToDatabase, self).__init__()
        self.database_path = database_path
        self.subject_information = subject_information
        self.measurement_information = measurement_information
        self.recording_information = recording_information
        self.stimuli_information = stimuli_information
        self.processes_list = processes_list
        if isinstance(include_waveforms, list):
            assert len(include_waveforms) == len(processes_list)
        self.include_waveforms = include_waveforms
        self.names_prefixes = names_prefixes
        self.check_duplicates = check_duplicates

    def run(self):
        start_time = time.time()
        _data_tables = []
        _all_statistical_methods = TestType().get_available_methods()
        statistical_methods = {}
        processing_tables = {}
        waveforms = pd.DataFrame()

        if self.names_prefixes is None:
            self.names_prefixes = [_process.name for _process in self.processes_list]
        assert len(self.names_prefixes) == len(self.processes_list)

        _include_waveforms = self.include_waveforms
        if isinstance(self.include_waveforms, bool):
            _include_waveforms = [self.include_waveforms] * len(self.processes_list)

        for _prefix, _process, _iwf in zip(self.names_prefixes, self.processes_list, _include_waveforms):
            # extract statistical tables
            for _test, _stats in _process.output_node.statistical_tests.items():
                _all_statistical_methods = statistical_methods.keys()
                if _test in _all_statistical_methods:
                    statistical_methods[_test] = pd.concat([statistical_methods[_test],
                                                            _stats.dropna(axis=1, how='all').copy()])
                else:
                    statistical_methods[_test] = _stats.copy()

            for _table_name, _table in _process.output_node.processing_tables_local.items():
                _all_processing_tables = processing_tables.keys()
                if _table_name in _all_processing_tables:
                    processing_tables[_table_name] = pd.concat([processing_tables[_table_name], _table.copy()])
                else:
                    processing_tables[_table_name] = _table.copy()

            if _iwf:
                _wave_forms = _process.output_node.data_to_pandas()
                _wave_forms['data_source'] = _prefix
                waveforms = pd.concat([waveforms, _wave_forms])

        for _test, _data in statistical_methods.items():
            _data_tables.append(PandasDataTable(table_name=_test,
                                                pandas_df=_data))
        for _table_name, _table in processing_tables.items():
            _data_tables.append(PandasDataTable(table_name=_table_name,
                                                pandas_df=_table))
        _data_tables.append(PandasDataTable(table_name='waveforms',
                                            pandas_df=waveforms))

        store_data(database_path=self.database_path,
                   subject_info=self.subject_information,
                   measurement_info=self.measurement_information,
                   recording_info=self.recording_information,
                   stimuli_info=self.stimuli_information,
                   check_duplicates=self.check_duplicates,
                   pandas_df=_data_tables)
        print("Saving took --- %s seconds ---" % (time.time() - start_time))



import peegy.io.readers.edf_bdf_reader as br
import matplotlib.pyplot as plt
import numpy as np
import pyedflib
__author__ = 'jundurraga'

f_name = '/run/media/jundurraga/Elements/Measurements/P300/Data/1_cijfers.edf'
header = br.read_edf_bdf_header(file_name=f_name)
ch = np.array([0, 1])
event_table = br.get_event_channel(header=header, ini_time=0, end_time=None,
                                   event_channel_label='EDF Annotations')
data, _events, _units, annotations = br.get_data(header=header, channels_idx=ch, ini_time=0, end_time=None)
fs = header['fs'][0]
x = np.arange(0, data.shape[0]) / fs
fig = plt.figure()
plt.plot(x, data - np.mean(data, axis=0))
plt.plot(x[_events.squeeze() > 0], _events[_events > 0], 'ro')

f = pyedflib.EdfReader(f_name)
f.readAnnotations()
n = f.signals_in_file
signal_labels = f.getSignalLabels()
sigbufs = np.zeros((f.getNSamples()[0], n))
for i in np.arange(n):
    sigbufs[:, i] = f.readSignal(i)
xx = np.arange(0, sigbufs.shape[0])/f.getSampleFrequency(0)
plt.figure()
plt.plot(xx, sigbufs[:, 0:n])
plt.show()
plt.show()

plt.figure()
f = np.arange(data.shape[0]) * fs / data.shape[0]
plt.plot(f, np.abs(np.fft.fft(data, axis=0)) * 2 / data.shape[0])
plt.xlim(0, 60)
plt.show()



"""
.. _tut-assr-ht2-test-sim:

##############################################
ASSR Hotelling-T2 test (Simulated)
##############################################

In this example we simulate an ASSR, and we assess its significance using the Hotelling's T2 test.
This method uses the real and imaginary part of the frequency bin of interest to assess whether the points from
each epoch are significantly different from zero (the center of the polar complex-plane)

"""
# Enable below for interactive backend
# import matplotlib
# if 'Qt5Agg' in matplotlib.rcsetup.all_backends:
#     matplotlib.use('Qt5Agg')
from peegy.processing.pipe.pipeline import PipePool
from peegy.processing.pipe.definitions import Events, Domain
from peegy.processing.pipe.general import ReferenceData, FilterData, AutoRemoveBadChannels
from peegy.processing.pipe.epochs import AverageEpochsFrequencyDomain, AverageEpochs, EpochData
from peegy.processing.pipe.attach import AppendGFPChannel
from peegy.processing.pipe.statistics import FTest
from peegy.processing.pipe.plot import PlotWaveforms, PlotTopographicMap
from peegy.processing.pipe.spatial_filtering import CreateAndApplySpatialFilter
from peegy.processing.pipe.simulate import GenerateInputData
from peegy.processing.tools.template_generator.auditory_waveforms import aep
from peegy.processing.pipe.storage import MeasurementInformation, SubjectInformation, SaveToDatabase
from peegy.io.storage.data_storage_reading_tools import sqlite_tables_to_pandas, sqlite_waveforms_to_pandas
from peegy.io.storage.plot_tools import plot_time_frequency_responses, plot_topographic_maps
import os
import astropy.units as u
import numpy as np


# %%
# Generate some data
# ===================
# We generate some auditory steady-state response (ASSR)


fs = 256.0 * u.Hz
epoch_length = 4.0 * u.s
epoch_length = np.ceil(epoch_length * fs) / fs  # fit to fs rate
assr_frequency = 41.0 * u.Hz
assr_frequency = np.ceil(epoch_length * assr_frequency) / epoch_length  # fit to epoch length

# here we pick some random frequencies to test statistical detection
random_frequencies = np.unique(np.random.rand(2)*5) * u.Hz

test_frequencies = np.concatenate((
    [assr_frequency,
     2 * assr_frequency],
    random_frequencies))

template_waveform, _ = aep(fs=fs)
n_channels = 32
event_times = np.arange(0, 360.0, 1 / assr_frequency.to(u.Hz).value)
reader = GenerateInputData(template_waveform=template_waveform,
                           fs=fs,
                           n_channels=n_channels,
                           mixing_matrix=np.diag(np.arange(n_channels)) / n_channels,
                           snr=0.05,
                           layout_file_name='biosemi32.lay',
                           event_times=event_times,
                           event_code=1.0,
                           figures_subset_folder='assr_ht2_test',
                           noise_seed=0)
reader.run()

# %%
# Resize events
# ===================
# Now we keep events at intervals that correspond to our desired epoch length


events = reader.output_node.events.get_events(code=1)
# skip events to preserve only those at each epoch point
_new_events = Events(events=events[0:-1:int(epoch_length * assr_frequency)])
reader.output_node.events = _new_events

# %%
# Start the pipeline
# ===================
# Some processing to obtain clean epochs

pipeline = PipePool()
pipeline['referenced'] = ReferenceData(reader,
                                       reference_channels=['Cz'],
                                       invert_polarity=True)
pipeline['channel_cleaned'] = AutoRemoveBadChannels(pipeline['referenced'])

pipeline['time_filtered_data'] = FilterData(pipeline['channel_cleaned'],
                                            high_pass=2.0 * u.Hz,
                                            low_pass=100.0 * u.Hz)
pipeline['time_epochs'] = EpochData(pipeline['time_filtered_data'],
                                    event_code=1.0,
                                    base_line_correction=False)
pipeline.run()

# %%
# Plot components in the frequency-domain
# ---------------------------------------
# Spatial filter is a applied in the frequency-domain

pipeline['dss_time_epochs'] = CreateAndApplySpatialFilter(pipeline['time_epochs'],
                                                          block_size=10,
                                                          return_figures=True,
                                                          projection_domain=Domain.frequency,
                                                          plot_y_lim=[0, 0.2])
pipeline.run()

# %%
# Average epochs
# ---------------------------------------
# We compute the average and simultaneously get statistical tests on the test_frequencies


pipeline['fft_ave'] = AverageEpochsFrequencyDomain(pipeline['dss_time_epochs'],
                                                   n_fft=int(epoch_length*fs),
                                                   weight_frequencies=assr_frequency,
                                                   test_frequencies=test_frequencies,
                                                   delta_frequency=2 * u.Hz)
pipeline['time_ave'] = AverageEpochs(pipeline['dss_time_epochs'])

pipeline['std_fft_ave'] = AverageEpochsFrequencyDomain(pipeline['time_epochs'],
                                                       n_fft=int(epoch_length * fs),
                                                       weight_frequencies=assr_frequency,
                                                       test_frequencies=test_frequencies,
                                                       delta_frequency=2 * u.Hz)
pipeline.run()

# %%
# Compute global field power (GFP)
# ---------------------------------------
# We compute the GFP across channels and epochs. A new channel with the GFP is appended to the data

pipeline['fft_ave_gfp'] = AppendGFPChannel(pipeline['fft_ave'])
pipeline.run()

pipeline['std_fft_ave_gfp'] = AppendGFPChannel(pipeline['std_fft_ave'])
pipeline.run()

pipeline['fft_ave_gfp_ftest'] = FTest(
    pipeline['fft_ave_gfp'],
    test_frequencies=test_frequencies,
    delta_frequency=5 * u.Hz)
pipeline.run()
pipeline['std_fft_ave_gfp_ftest'] = FTest(
    pipeline['std_fft_ave_gfp'],
    test_frequencies=test_frequencies,
    delta_frequency=5 * u.Hz)
pipeline.run()

pipeline['fft_ave'].output_node.statistical_tests['hotelling_t2_freq'][
    ["test_name", "df_1", "df_2", "f", "f_critic", "p_value"]].head()
pipeline.run()

# %%
# Generate figures
# ---------------------------------------
# Now we run plot the average waveforms and show the stats


pipeline['waveform_plotter'] = PlotWaveforms(pipeline['fft_ave_gfp_ftest'],
                                             overlay=[pipeline['std_fft_ave_gfp_ftest']],
                                             plot_x_lim=[0, 90],
                                             ch_to_plot=np.array(['O2', 'T8', 'T7', 'GFP']),
                                             statistical_test='f_test_freq',
                                             show_following_stats=['f'],
                                             return_figures=True,
                                             user_naming_rule='standard_and_dss')
pipeline.run()
pipeline['topographic_map'] = PlotTopographicMap(pipeline['fft_ave'],
                                                 topographic_channels=np.array(['O2', 'T8', 'T7', 'GFP']),
                                                 plot_x_lim=[0, 90],
                                                 plot_y_lim=[0, 6],
                                                 return_figures=True)
pipeline.run()

# %%
# Save results to a database
# ---------------------------------------
# We get the measurements we are interested in and save them into a database

subject_info = SubjectInformation(subject_id='Test_Subject')
measurement_info = MeasurementInformation(
    date='Today',
    experiment='sim')

_parameters = {'Type': 'ASSR'}
database_path = reader.input_node.paths.test_path.joinpath('assr_ht2_test.sqlite')
pipeline['database'] = SaveToDatabase(database_path=database_path,
                                      measurement_information=measurement_info,
                                      subject_information=subject_info,
                                      recording_information={'recording_device': 'dummy_device'},
                                      stimuli_information=_parameters,
                                      processes_list=[pipeline['fft_ave_gfp'],
                                                      pipeline['std_fft_ave_gfp'],
                                                      pipeline['time_ave']],
                                      include_waveforms=True
                                      )
pipeline.run()

# %%
# Generate pipeline diagram
# ------------------------------------
pipeline.diagram(file_name=reader.output_node.paths.figures_current_dir + 'pipeline.png',
                 return_figure=True,
                 dpi=600)


# Read waveforms from generated database and plot them
# ------------------------------------
df_waves = sqlite_waveforms_to_pandas(database_path=database_path,
                                      group_factors=['data_source', 'channel'],
                                      channels=['T8', 'T7'])

fig_out_1 = plot_time_frequency_responses(dataframe=df_waves,
                                          rows_by='channel',
                                          cols_by='data_source',
                                          title_by='col',
                                          show_legend=True)

fig_out_2 = plot_time_frequency_responses(dataframe=df_waves,
                                          rows_by='channel',
                                          cols_by='domain',
                                          title_by='both',
                                          show_legend=True)

fig_out_3 = plot_time_frequency_responses(dataframe=df_waves,
                                          rows_by='channel',
                                          cols_by='domain',
                                          title_by='both',
                                          show_legend=True)

df_ht2 = sqlite_tables_to_pandas(database_path=database_path,
                                 tables=['hotelling_t2_freq'])['hotelling_t2_freq']

# We show the topographic maps for the two Hotelling T2 (standard processing and denoise using dss)

fig_out_4 = plot_topographic_maps(dataframe=df_ht2,
                                  channels_column='channel',
                                  rows_by='data_source',
                                  layout='biosemi32.lay',
                                  color_map_label='Amplitude [uV]',
                                  topographic_value='mean_amplitude')
